import sys
import weakref

from chid cimport *
from libc.stddef cimport wchar_t, size_t


__version__ = "0.15.0"


hid_init()


cdef extern from "<ctype.h>":
    int wcslen(wchar_t*)

cdef extern from "<stdlib.h>":
    void free(void* ptr)
    void* malloc(size_t size)

cdef extern from *:
    object PyUnicode_FromWideChar(const wchar_t *w, Py_ssize_t size)
    Py_ssize_t PyUnicode_AsWideChar(object unicode, wchar_t *w, Py_ssize_t size)

cdef object U(wchar_t *wcs):
    if wcs == NULL:
        return ''
    cdef int n = wcslen(wcs)
    return PyUnicode_FromWideChar(wcs, n)

def enumerate(int vendor_id=0, int product_id=0):
    """Return a list of discovered HID devices.

    The fields of dict are:

     - 'path'
     - 'vendor_id'
     - 'product_id'
     - 'serial_number'
     - 'release_number'
     - 'manufacturer_string'
     - 'product_string'
     - 'usage_page'
     - 'usage'
     - 'interface_number'

    :param vendor_id: Vendor id to look for, default = 0
    :type vendor_id: int, optional
    :param product_id: Product id to look for, default = 0
    :type product_id: int, optional
    :return: List of device dictionaries
    :rtype: List[Dict]
    """
    cdef hid_device_info* info
    with nogil:
        info = hid_enumerate(vendor_id, product_id)
    cdef hid_device_info* c = info
    res = []
    while c:
        res.append({
            'path': c.path,
            'vendor_id': c.vendor_id,
            'product_id': c.product_id,
            'serial_number': U(c.serial_number),
            'release_number': c.release_number,
            'manufacturer_string': U(c.manufacturer_string),
            'product_string': U(c.product_string),
            'usage_page': c.usage_page,
            'usage': c.usage,
            'interface_number': c.interface_number,
            'bus_type': c.bus_type,
        })
        c = c.next
    hid_free_enumeration(info)
    return res

def version_str():
    """Return a runtime version string of the hidapi C library.

    :return: version string of library
    :rtype: str
    """
    return (<bytes>hid_version_str()).decode('ascii')

cdef class _closer:
    """Wrap a hid_device *ptr and a provide a way to call hid_close() on it.

    Used internally for weakref.finalize, which only accepts Python objects.
    """

    cdef hid_device *_ptr

    @staticmethod
    cdef wrap(hid_device *ptr):
        cdef _closer closer = _closer()
        closer._ptr = ptr
        return closer

    def close(self):
        hid_close(self._ptr)

cdef class device:
    """Device class.

    A device instance can be used to read from and write to a HID device.
    """

    cdef hid_device *_c_hid
    cdef object __weakref__  # enable weak-reference support
    cdef object _close

    def open(self, int vendor_id=0, int product_id=0, unicode serial_number=None):
        """Open the connection.

        :param vendor_id: Vendor id to connect to, default = 0
        :type vendor_id: int, optional
        :param product_id: Product id to connect to, default = 0
        :type product_id: int, optional
        :param serial_number:
        :type serial_number: unicode, optional
        :raises IOError:
        """
        cdef wchar_t * cserial_number = NULL
        cdef int serial_len
        cdef Py_ssize_t result
        if self._c_hid != NULL:
            raise RuntimeError('already open')
        try:
            if serial_number is not None:
                serial_len = len(serial_number)
                cserial_number = <wchar_t*>malloc(sizeof(wchar_t) * (serial_len+1))
                if cserial_number == NULL:
                    raise MemoryError()
                result = PyUnicode_AsWideChar(serial_number, cserial_number, serial_len)
                if result == -1:
                    raise ValueError("invalid serial number string")
                cserial_number[serial_len] = 0  # Must explicitly null-terminate
            self._c_hid = hid_open(vendor_id, product_id, cserial_number)
        finally:
            if cserial_number != NULL:
                free(cserial_number)
        if self._c_hid == NULL:
            raise IOError('open failed')
        self._close = weakref.finalize(self, _closer.wrap(self._c_hid).close)

    def open_path(self, bytes path):
        """Open connection by path.

        :param path: Path to device
        :type path: bytes
        :raises IOError:
        """
        cdef char* cbuff = path
        if self._c_hid != NULL:
            raise RuntimeError('already open')
        self._c_hid = hid_open_path(cbuff)
        if self._c_hid == NULL:
            raise IOError('open failed')
        self._close = weakref.finalize(self, _closer.wrap(self._c_hid).close)

    def close(self):
        """Close connection.

        This should always be called after opening a connection.
        """
        if self._c_hid != NULL:
            self._close()
            self._close.detach()
            self._c_hid = NULL

    def write(self, buff):
        """Accept a list of integers (0-255) and send them to the device.

        :param buff: Data to write (must be convertible to `bytes`)
        :type buff: Any
        :return: Write result
        :rtype: int
        """
        if self._c_hid == NULL:
            raise ValueError('not open')
        # convert to bytes
        if sys.version_info < (3, 0):
            buff = ''.join(map(chr, buff))
        else:
            buff = bytes(buff)
        cdef hid_device * c_hid = self._c_hid
        cdef unsigned char* cbuff = buff # covert to c string
        cdef size_t c_buff_len = len(buff)
        cdef int result
        with nogil:
            result = hid_write(c_hid, cbuff, c_buff_len)
        return result

    def set_nonblocking(self, int v):
        """Set the nonblocking flag.

        :param v: Flag value (1 or 0, True or False)
        :type v: int, bool
        :return: Flag result
        :rtype: int
        """
        if self._c_hid == NULL:
            raise ValueError('not open')
        return hid_set_nonblocking(self._c_hid, v)

    def read(self, int max_length, int timeout_ms=0):
        """Return a list of integers (0-255) from the device up to max_length bytes.

        :param max_length: Maximum number of bytes to read
        :type max_length: int
        :param timeout_ms: Number of milliseconds until timeout (default: no timeout)
        :type timeout_ms: int, optional
        :return: Read bytes
        :rtype: List[int]
        """
        if self._c_hid == NULL:
            raise ValueError('not open')
        cdef unsigned char lbuff[16]
        cdef unsigned char* cbuff
        cdef size_t c_max_length = max_length
        cdef int c_timeout_ms = timeout_ms
        cdef hid_device * c_hid = self._c_hid
        try:
            if max_length <= 16:
                cbuff = lbuff
            else:
                cbuff = <unsigned char *>malloc(max_length)
            if timeout_ms > 0:
                with nogil:
                    n = hid_read_timeout(c_hid, cbuff, c_max_length, c_timeout_ms)
            else:
                with nogil:
                    n = hid_read(c_hid, cbuff, c_max_length)
            if n == -1:
                raise IOError('read error')
            res = list(cbuff[:n])
        finally:
            if max_length > 16:
                free(cbuff)
        return res

    def get_manufacturer_string(self):
        """Return manufacturer string (e.g. vendor name).

        :return:
        :rtype: str
        :raises ValueError: If connection is not opened.
        :raises IOError:
        """
        if self._c_hid == NULL:
            raise ValueError('not open')
        cdef wchar_t buff[255]
        cdef int r = hid_get_manufacturer_string(self._c_hid, buff, 255)
        if r < 0:
            raise IOError('get manufacturer string error')
        return U(buff)


    def get_product_string(self):
        """Return product string (e.g. device description).

        :return:
        :rtype: str
        :raises ValueError: If connection is not opened.
        :raises IOError:
        """
        if self._c_hid == NULL:
            raise ValueError('not open')
        cdef wchar_t buff[255]
        cdef int r = hid_get_product_string(self._c_hid, buff, 255)
        if r < 0:
            raise IOError('get product string error')
        return U(buff)

    def get_serial_number_string(self):
        """Return serial number.

        :return:
        :rtype: str
        :raises ValueError: If connection is not opened.
        :raises IOError:
        """
        if self._c_hid == NULL:
            raise ValueError('not open')
        cdef wchar_t buff[255]
        cdef int r = hid_get_serial_number_string(self._c_hid, buff, 255)
        if r < 0:
            raise IOError('get serial number string error')
        return U(buff)

    def get_indexed_string(self, index):
        """Return indexed string.

        :return:
        :rtype: str
        :raises ValueError: If connection is not opened.
        :raises IOError:
        """
        if self._c_hid == NULL:
            raise ValueError('not open')
        cdef wchar_t buff[255]
        cdef unsigned char c_index = index
        cdef int r = hid_get_indexed_string(self._c_hid, c_index, buff, 255)
        if r < 0:
            raise IOError('get indexed string error')
        return U(buff)

    def send_feature_report(self, buff):
        """Accept a list of integers (0-255) and send them to the device.

        :param buff: Data to send (must be convertible into bytes)
        :type buff: any
        :return: Send result
        :rtype: int
        """
        if self._c_hid == NULL:
            raise ValueError('not open')
        # convert to bytes
        if sys.version_info < (3, 0):
            buff = ''.join(map(chr, buff))
        else:
            buff = bytes(buff)
        cdef hid_device * c_hid = self._c_hid
        cdef unsigned char* cbuff = buff  # covert to c string
        cdef size_t c_buff_len = len(buff)
        cdef int result
        with nogil:
            result = hid_send_feature_report(c_hid, cbuff, c_buff_len)
        return result

    def get_report_descriptor(self):
        """Return the HID Report Descriptor for this device.

        :return:
        :rtype: List[int]
        :raises ValueError: If connection is not opened.
        :raises IOError: If read error
        """
        if self._c_hid == NULL:
            raise ValueError('not open')

        cdef unsigned char cbuff[HID_API_MAX_REPORT_DESCRIPTOR_SIZE]
        cdef hid_device * c_hid = self._c_hid
        cdef int n
        with nogil:
            n = hid_get_report_descriptor(c_hid, cbuff, sizeof(cbuff))
        if n < 0:
            raise IOError('read error')
        return list(cbuff[:n])

    def get_feature_report(self, int report_num, int max_length):
        """Receive feature report.

        :param report_num:
        :type report_num: int
        :param max_length:
        :type max_length: int
        :return: Incoming feature report
        :rtype: List[int]
        :raises ValueError: If connection is not opened.
        :raises IOError:
        """
        if self._c_hid == NULL:
            raise ValueError('not open')
        cdef hid_device * c_hid = self._c_hid
        cdef unsigned char lbuff[16]
        cdef unsigned char* cbuff
        cdef size_t c_max_length = max_length
        cdef int n
        try:
            if max_length <= 16:
                cbuff = lbuff
            else:
                cbuff = <unsigned char *>malloc(max_length)
            cbuff[0] = report_num
            with nogil:
                n = hid_get_feature_report(c_hid, cbuff, c_max_length)
            if n < 0:
                raise IOError('read error')
            res = list(cbuff[:n])
        finally:
            if max_length > 16:
                free(cbuff)
        return res

    def get_input_report(self, int report_num, int max_length):
        """Get input report

        :param report_num:
        :type report_num: int
        :param max_length:
        :type max_length: int
        :return:
        :rtype: List[int]
        :raises ValueError: If connection is not opened.
        :raises IOError:
        """
        if self._c_hid == NULL:
            raise ValueError('not open')
        cdef hid_device * c_hid = self._c_hid
        cdef unsigned char lbuff[16]
        cdef unsigned char* cbuff
        cdef size_t c_max_length = max_length
        cdef int n
        try:
            if max_length <= 16:
                cbuff = lbuff
            else:
                cbuff = <unsigned char *>malloc(max_length)
            cbuff[0] = report_num
            with nogil:
                n = hid_get_input_report(c_hid, cbuff, c_max_length)
            if n < 0:
                raise IOError('read error')
            res = list(cbuff[:n])
        finally:
            if max_length > 16:
                free(cbuff)
        return res

    def error(self):
        """Get error from device, or global error if no device is opened.

        :return:
        :rtype: str
        :raises IOError:
        """
        return U(<wchar_t*>hid_error(self._c_hid))


# Finalize the HIDAPI library *only* once there are no more references to this
# module, and it is being garbage collected.
weakref.finalize(sys.modules[__name__], hid_exit)



from libc.stddef cimport wchar_t, size_t

cdef extern from "<hidapi.h>":
    const int HID_API_MAX_REPORT_DESCRIPTOR_SIZE

    ctypedef struct hid_device:
        pass

    ctypedef enum hid_bus_type:
        HID_API_BUS_UNKNOWN = 0x00,
        HID_API_BUS_USB = 0x01,
        HID_API_BUS_BLUETOOTH = 0x02,
        HID_API_BUS_I2C = 0x03,
        HID_API_BUS_SPI = 0x04

    cdef struct hid_device_info:
        char *path
        unsigned short vendor_id
        unsigned short product_id
        wchar_t *serial_number
        unsigned short release_number
        wchar_t *manufacturer_string
        wchar_t *product_string
        unsigned short usage_page
        unsigned short usage
        int interface_number
        hid_device_info *next
        hid_bus_type bus_type

    hid_device_info* hid_enumerate(unsigned short, unsigned short) nogil
    void hid_free_enumeration(hid_device_info*)

    hid_device* hid_open(unsigned short, unsigned short, const wchar_t*)
    hid_device* hid_open_path(char *path)
    void hid_close(hid_device *)
    int hid_init()
    int hid_exit()
    int hid_write(hid_device* device, unsigned char *data, int length) nogil
    int hid_read(hid_device* device, unsigned char* data, int max_length) nogil
    int hid_read_timeout(hid_device* device, unsigned char* data, int max_length, int milliseconds) nogil
    int hid_set_nonblocking(hid_device* device, int value)
    int hid_get_report_descriptor(hid_device* device, unsigned char *data, int length) nogil
    int hid_send_feature_report(hid_device* device, unsigned char *data, int length) nogil
    int hid_get_feature_report(hid_device* device, unsigned char *data, int length) nogil
    int hid_get_input_report(hid_device* device, unsigned char *data, int length) nogil

    int hid_get_manufacturer_string(hid_device*, wchar_t *, size_t)
    int hid_get_product_string(hid_device*, wchar_t *, size_t)
    int hid_get_serial_number_string(hid_device*, wchar_t *, size_t)
    int hid_get_indexed_string(hid_device*, int, wchar_t *, size_t)
    wchar_t *hid_error(hid_device *)
    char* hid_version_str()



#!/usr/bin/python
from Cython.Build import cythonize
from setuptools import setup, Extension, Distribution
from subprocess import call, PIPE, Popen
import os
import re
import shlex
import subprocess
import sys

min_required_hidapi_version = "0.14"

libusb_pkgconfig = "libusb-1.0 >= 1.0.9"
hidapi_libusb_pkgconfig = "hidapi-libusb >= " + min_required_hidapi_version
hidapi_hidraw_pkgconfig = "hidapi-hidraw >= " + min_required_hidapi_version
hidapi_pkgconfig = "hidapi >= " + min_required_hidapi_version

tld = os.path.abspath(os.path.dirname(__file__))
embedded_hidapi_topdir = os.path.join(tld, "hidapi")
embedded_hidapi_include = os.path.join(embedded_hidapi_topdir, "hidapi")


def to_bool(bool_str):
    if str(bool_str).lower() in ["true", "t", "1", "on", "yes", "y"]:
        return True
    return False


def get_extension_compiler_type():
    """
    Returns a compiler to be used by setuptools to build Extensions
    Taken from https://github.com/pypa/setuptools/issues/2806#issuecomment-961805789
    """
    d = Distribution()
    build_ext = Distribution().get_command_obj("build_ext")
    build_ext.finalize_options()
    # register an extension to ensure a compiler is created
    build_ext.extensions = [Extension("ignored", ["ignored.c"])]
    # disable building fake extensions
    build_ext.build_extensions = lambda: None
    # run to populate self.compiler
    build_ext.run()
    return build_ext.compiler.compiler_type


# this could have been just pkgconfig.configure_extension(), but: https://github.com/matze/pkgconfig/issues/65
# additionally contains a few small improvements
def pkgconfig_configure_extension(ext, package):
    pkg_config_exe = os.environ.get("PKG_CONFIG", None) or "pkg-config"

    def exists(package):
        cmd = f"{pkg_config_exe} --exists '{package}'"
        return call(shlex.split(cmd)) == 0

    if not exists(package):
        raise Exception(f"pkg-config package '{package}' not found")

    def query_pkg_config(package, *options):
        cmd = f"{pkg_config_exe} {' '.join(options)} '{package}'"
        proc = Popen(shlex.split(cmd), stdout=PIPE, stderr=PIPE)
        out, err = proc.communicate()

        return out.rstrip().decode("utf-8")

    def query_and_extend(option, target):
        os_opts = ["--msvc-syntax"] if get_extension_compiler_type() == "msvc" else []
        flags = query_pkg_config(package, *os_opts, option)
        flags = flags.replace('\\"', "")
        if flags:
            target.extend(re.split(r"(?<!\\) ", flags))

    query_and_extend("--cflags", ext.extra_compile_args)
    query_and_extend("--libs", ext.extra_link_args)

    return ext


def hidapi_src(platform):
    return os.path.join(embedded_hidapi_topdir, platform, "hid.c")


def hid_from_embedded_hidapi():
    # TODO: what about MinGW/msys?
    if sys.platform.startswith("win") or sys.platform.startswith("cygwin"):
        modules = [
            Extension(
                "hid",
                sources=["hid.pyx", hidapi_src("windows")],
                include_dirs=[embedded_hidapi_include],
                extra_compile_args=["-DHID_API_NO_EXPORT_DEFINE"],
                libraries=["setupapi"],
            )
        ]

    elif sys.platform.startswith("darwin"):
        macos_sdk_path = (
            subprocess.check_output(["xcrun", "--show-sdk-path"]).decode().strip()
        )
        modules = [
            Extension(
                "hid",
                sources=["hid.pyx", hidapi_src("mac")],
                include_dirs=[embedded_hidapi_include],
                # TODO: remove -Wno-unreachable-code when the time comes: https://github.com/cython/cython/issues/3172
                extra_compile_args=[
                    "-isysroot",
                    macos_sdk_path,
                    "-Wno-unreachable-code",
                ],
                # TODO: remove '-framework AppKit' after switching to 0.14.1 or newer
                extra_link_args=[
                    "-framework",
                    "IOKit",
                    "-framework",
                    "CoreFoundation",
                    "-framework",
                    "AppKit",
                ],
            )
        ]

    elif sys.platform.startswith("linux"):
        modules = []
        if "--with-hidraw" in sys.argv:
            sys.argv.remove("--with-hidraw")
            HIDAPI_WITH_HIDRAW = True
        else:
            HIDAPI_WITH_HIDRAW = to_bool(os.getenv("HIDAPI_WITH_HIDRAW"))
        if "--with-libusb" in sys.argv:
            sys.argv.remove("--with-libusb")
            HIDAPI_WITH_LIBUSB = True
        else:
            HIDAPI_WITH_LIBUSB = to_bool(os.getenv("HIDAPI_WITH_LIBUSB"))

        # make libusb backend default if none is specified
        if not HIDAPI_WITH_HIDRAW and not HIDAPI_WITH_LIBUSB:
            HIDAPI_WITH_LIBUSB = True

        if HIDAPI_WITH_LIBUSB:
            hidraw_module = "hidraw"
            modules.append(
                pkgconfig_configure_extension(
                    Extension(
                        "hid",
                        sources=["hid.pyx", hidapi_src("libusb")],
                        include_dirs=[embedded_hidapi_include],
                    ),
                    libusb_pkgconfig,
                )
            )
        elif HIDAPI_WITH_HIDRAW:
            hidraw_module = "hid"
        else:
            raise ValueError("Unknown HIDAPI backend")

        modules.append(
            Extension(
                hidraw_module,
                sources=["hidraw.pyx", hidapi_src("linux")],
                include_dirs=[embedded_hidapi_include],
                libraries=["udev"],
            )
        )

    else:
        modules = [
            pkgconfig_configure_extension(
                Extension(
                    "hid",
                    sources=["hid.pyx", hidapi_src("libusb")],
                    include_dirs=[embedded_hidapi_include],
                ),
                libusb_pkgconfig,
            )
        ]

    return modules


def hid_from_system_hidapi():
    if sys.platform.startswith("linux"):
        modules = []
        if "--with-hidraw" in sys.argv:
            sys.argv.remove("--with-hidraw")
            HIDAPI_WITH_HIDRAW = True
        else:
            HIDAPI_WITH_HIDRAW = to_bool(os.getenv("HIDAPI_WITH_HIDRAW"))
        if "--with-libusb" in sys.argv:
            sys.argv.remove("--with-libusb")
            HIDAPI_WITH_LIBUSB = True
        else:
            HIDAPI_WITH_LIBUSB = to_bool(os.getenv("HIDAPI_WITH_LIBUSB"))

        # make libusb backend default if none is specified
        if not HIDAPI_WITH_HIDRAW and not HIDAPI_WITH_LIBUSB:
            HIDAPI_WITH_LIBUSB = True

        if HIDAPI_WITH_LIBUSB:
            hidraw_module = "hidraw"
            modules.append(
                pkgconfig_configure_extension(
                    Extension("hid", sources=["hid.pyx"]), hidapi_libusb_pkgconfig
                )
            )
        elif HIDAPI_WITH_HIDRAW:
            hidraw_module = "hid"
        else:
            raise ValueError("Unknown HIDAPI backend")

        modules.append(
            pkgconfig_configure_extension(
                Extension(hidraw_module, sources=["hidraw.pyx"]),
                hidapi_hidraw_pkgconfig,
            )
        )
    else:
        modules = [
            pkgconfig_configure_extension(
                Extension("hid", sources=["hid.pyx"]), hidapi_pkgconfig
            )
        ]

    return modules


def find_version():
    filename = os.path.join(tld, "hid.pyx")
    with open(filename) as f:
        text = f.read()
    match = re.search(r"^__version__ = \"(.*)\"$", text, re.MULTILINE)
    if not match:
        raise RuntimeError("cannot find version")
    return match.group(1)


if "--with-system-hidapi" in sys.argv:
    sys.argv.remove("--with-system-hidapi")
    HIDAPI_SYSTEM_HIDAPI = True
else:
    HIDAPI_SYSTEM_HIDAPI = to_bool(os.getenv("HIDAPI_SYSTEM_HIDAPI"))

if HIDAPI_SYSTEM_HIDAPI:
    modules = hid_from_system_hidapi()
else:
    modules = hid_from_embedded_hidapi()


setup(
    name="hidapi",
    version=find_version(),
    description="A Cython interface to the hidapi from https://github.com/libusb/hidapi",
    long_description=open("README.rst", "rt").read(),
    author="Pavol Rusnak",
    author_email="pavol@rusnak.io",
    maintainer="Pavol Rusnak",
    maintainer_email="pavol@rusnak.io",
    url="https://github.com/trezor/cython-hidapi",
    package_dir={"hid": "hidapi/*"},
    classifiers=[
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "License :: OSI Approved :: BSD License",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: 3.14",
    ],
    ext_modules=cythonize(modules, language_level=3),
    setup_requires=["setuptools>=19.0"],
)
